/*->c.undo */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "flex.h"

#include "err.h"
#include "wos.h"
#include "fsx.h"
#include "xext.h"
#include "poll.h"
#include "alloc.h"
#include "transform.h"
#include "xprocess.h"
#include "deb.h"


#include "etc.h"

#include "trans.h"

#include "constants.h"
#include "str.h"
#include "reslink.h"

#include "file.h"
#include "view.h"

#include "tools.h"


#include "undo.h"





typedef struct undostr
{
 int          key;
 int          type;
 chainstr     chain;
 flagstr      flags;

 int          ustart;
 int          uend;
 int          ulast;
 int          uptr;
 int          uostart;

 int          rstart;
 int          rend;
 int          rlast;
 int          rptr;
 int          rostart;

 int          count;

 unsigned int state:8;
 unsigned int rsbyte:1;
 unsigned int rebyte:1;
 unsigned int usbyte:1;
 unsigned int uebyte:1;
 unsigned int rscrap:1;
 unsigned int uscrap:1;
 unsigned int uewbyte:1;
 unsigned int rswbyte:1;
 unsigned int uwfn:1;
 unsigned int rwfn:1;

 int          maxsize;

 int          spare[8];

 char         data[1];
} undostr;



/*

A general undo opcode 

0  length

4  opcode
8  data
   length    (offset-length)->opcode
...
  opcode
  data
  length     (offset-length)->opcode
...


Always starts with a 0 byte on a word boundary
Last opcode in a sequence has high bit set
Transition to byte opcodes is marked by a length with the high bit set


Byte opcodes

start and terminate with a 0 byte

Pass on to short op-code handler(s)

Pad to word boundaries with 0 bytes

Undo buffer

------  maxsize 

      <- rend

      <-- rostart
      <-- rlast
      <-- rptr

      <- rstart


~~~~~

      <- uend

      <-- uptr      - pointer to top of undo stack
      <-- ulast     - pointer to last opcode
      <-- uostart   - pointer to start of this undo sequence



      <- ustart

------ 0

state is either STNONE - doing nothing
                STNEW  - writing new info, write to undo, redo trashed
                STUNDO - undoing, read from undo, write to redo
                STREDO - redoing, write to undo, read from redo

STNEW  Undo data expands up in memory

STUNDO Redo data expands down, undo contracts down

STREDO Redo data contracts up, undo expands up

We have to ensure that the memory ustart-uend and rstart-rend is all that
is in use. Also ensure that rostart and uostart denote operation starts.

uptr is an ascending, empty stack
rptr is a  descending, full stack

rstart and ustart are full  descending
rend   and uend   are empty ascending


*/


/* 

   there seems to be an oddity,
   int that one big write, cannot be read back in a number of steps
   or else order gets swopped

   this is caused by the redo stack which goes down in memory,
   therefore data is stored in blocks going down, whilst their contents
   go up.

   the only way around this is to store stuff going down, but doing so
   will affect performance


   XMEMCPYX * rXMEMCPYX(char *s1,char *s2,int n)
   {
    while(n--) *s1--=*s2++;                                                        }


   we assume that this ability to break writes down, only matters
   for atomic scrap functions, and provide a function that does one write
   int two pieces.

*/





#define UNDOXLAST 0x80000000
#define UNDOXBYTE 0x80000000

/* max length of an undo opcode name */

#define MAXUNDOSTRING 64

/* always have this much space available above uend etc. */

#define MINUNDOSPACE  32

/* amount of byte codes to dispose of in one go */

#define BYTECHUNK     32

/* initial space between rstart and uend */

#define REDOGAP       512

/* delft. size of undo buffer */

#define DEFUNDOSIZE 0x400


#define STNONE       0
#define STNEW        1
#define STUNDO       2
#define STREDO       3
#define STUNDOSCRAP  4
#define STREDOSCRAP  5
#define STFASTSCRAP  6
#define STCHANGE     7



#define UNDOCHUNK (12*sizeof(undofnstr))



int saveundo;     /* save undo buffer ?       */
int undosize;     /* default undo buffer size */



static undofnstr * undos;
static int         undon;

static undosbfn      undosb;
static undosbnamefn  undosbname;


/* variables used when stuff is being scrapped as written */

static char * fastptr;
static int    fastmax;





static int undocmp(const void * p1,const void * p2)
{
 return(((undofnstr*)p1)->tag-((undofnstr*)p2)->tag);
}


static os_error * undoexecute(int op,filestr * file,int scrap)
{
 os_error  * err;
 undofnstr * u;
 undofnstr   t;

 t.tag=op;

 u=(undofnstr*)bsearch(&t,undos,undon,sizeof(undofnstr),undocmp);

 if(u)
 {
  if(scrap)
  {
   if(u->scrap)
   {
    err=u->scrap(file);
   }
   else         err=NULL;
  }
  else
  {
   if(u->undo)
   {
    err=u->undo(file);
   }
   else          err=NULL;
  }
 }
 else  err=geterror(EU00);

 return(err);
}


static void getopstring(int op,char * p)
{
 undofnstr * u;
 undofnstr   t;

 t.tag=op;
 u=(undofnstr*)bsearch(&t,undos,undon,sizeof(undofnstr),undocmp);
 if(u && u->name) strcpy(p,u->name);
}



static int setptr(undostr * undop,int offset)
{
 if(undop->maxsize)
 {
  while(offset<0)               offset+=undop->maxsize;
  while(offset>=undop->maxsize) offset-=undop->maxsize;
 }
 return(offset);
}

static int roundup(undostr * undop,int ptr)
{
 return(setptr(undop,(ptr+0x3) & (~0x3))); /* rounds up */
}


static int rounddown(undostr * undop,int ptr)
{
 return(setptr(undop,ptr & (~0x3))); /* rounds down */
}



/* return (off1-off2) which should be +ve */

static int diffp(undostr * undop,int off1,int off2)
{
 off1=off1-off2;
 if(off1<0) off1+=undop->maxsize;
 if(off1>undop->maxsize) off1-=undop->maxsize;
 return(off1);
}


static int readbyte(undostr * undop,int offset)
{
 offset=setptr(undop,offset);
 return(undop->data[offset]);
}


static void writebyte(undostr * undop,int offset,int c)
{
 offset=setptr(undop,offset);
 undop->data[offset]=(char)c;
}


static int readword(undostr * undop,int offset)
{
 offset=setptr(undop,offset);
 return(*((int*)(undop->data+offset)));
}


static void writeword(undostr * undop,int offset,int c)
{
 offset=setptr(undop,offset);
 *((int*)(undop->data+offset))=c;
}


static void orrword(undostr * undop,int offset,int bits)
{
 int * p;
 offset=setptr(undop,offset);
 p=(int*)(undop->data+offset);
 *p=(*p)|bits;
}


static void bicword(undostr * undop,int offset,int bits)
{
 int * p;
 offset=setptr(undop,offset);
 p=(int*)(undop->data+offset);
 *p=(*p) & (~bits); 
}


static int getxword(undostr * undop,int offset,int bits)
{
 offset=setptr(undop,offset);
 return((*((int*)(undop->data+offset))) & (~bits));
}




/* called when a file is created to set up the undo buffer for it */

os_error * newundobuffer(filestr * file)
{
 os_error      * err;
 int             undo;
 undostr       * undop;
 int             initsize;

 initsize=undosize;

 err=NULL;


 if(!file->hdr.undo)
 {
  err=fob_create(initsize+sizeof(undostr)-sizeof(int),
                                                 file,OUNDO,(fobjptr)&undop);
  if(!err)
  {
   undo=undop->key;
   file->hdr.undo=undo;

   undop=(undostr*)fob_find(undo,file);
   undop->rstart=undop->rend=undop->rlast=0;
   undop->ustart=undop->uend=undop->ulast=0;
   undop->state=STNONE;
   undop->maxsize=initsize;
  }
 }
 return(err);
}



static undostr * getundo(filestr * file)
{
 int             undo;
 undostr       * undop;

 undo=file->hdr.undo;
 if(undo) undop=(undostr*)fob_find(undo,file);
 else     undop=NULL;
 return(undop);
}




#ifdef NEVER


static XMEMCPYX saveundobuffer(filestr * file)
{
 undostr * undop;

 undop=getundo(file);
 fs_saveblock("xx1",0xFFD,undop->data,undop->maxsize);
}



XMEMCPYX undodebug(XMEMCPYX)
{
 undostr       * undop;
 int             ptr;
 FILE          * fp;
 int             len;
 int             op;
 int             last;
 int             count;
 int             scount;
 int             lptr;

 count=20;


 fp=fopen("xx","wb");
 if(fp)
 {
  undop=getundo(menufile);

  ptr=undop->ustart;

  fprintf(fp,"ustart at %d uend at %d\n",ptr,undop->uend);

  while(ptr!=undop->uend && count--)
  {
   len=readword(undop,ptr);
   op=readword(undop,ptr+sizeof(int));

   fprintf(fp,"op %x length=%d at %d\n",op,len & (~UNDOXBYTE),ptr);
   ptr=setptr(undop,ptr+len);


   lptr=ptr;
   scount=10;

   while(scount--)
   {
    lptr=setptr(undop,lptr-sizeof(int));
    len=readword(undop,lptr);
    lptr=setptr(undop,lptr-len);
    op=readword(undop,lptr);

    fprintf(fp,"op %x len=%d at %d\n",op,len & (~UNDOXBYTE),lptr);

    if(op & UNDOXLAST) break;
   }


   fprintf(fp,"\n");
  }

  fclose(fp);
 }

 saveundobuffer(menufile);
}


#endif




/* can be used to find if undo code is going forward in time */

int undoforward(filestr * file)
{
 undostr       * undop;
 undop=getundo(file);
 return(undop->state==STNEW || undop->state==STREDO);
}



int undoactive(filestr * file)
{
 undostr       * undop;
 undop=getundo(file);
 return(undop->state!=STNEW && undop->state!=STNONE);
}



/* drop an item from the start of the undo queue */

static os_error * undodrop(filestr * file,undostr ** undopp)
{
 os_error * err;
 undostr  * undop;
 int        len;
 int        op;
 int        last;
 int        ptr;
 int        xptr;
 int        sptr;
 int        lptr;
 int        sstate;
 int        i;
 int        c;

/* dprintf(0,"undo drop");  */

 undop=*undopp;

 if(undop->usbyte)
 {
  ptr=undop->ustart;
  xptr=undop->uend;

  for(i=0;i<BYTECHUNK;i++)
  {
   c=undop->data[ptr];      /*have to inc ptr at least one byte */
   ptr=setptr(undop,ptr+1); /* cf case of 0bbb follwed by 0000  */
   if(!c) break;            /* roundup below won't skip 0000's  */
   if(ptr==xptr) break;
  }

  if(i<BYTECHUNK)
  {
   if(ptr==xptr)
   {
    xptr=ptr=roundup(undop,ptr);
   }
   else
   {
    ptr=roundup(undop,ptr);
    bicword(undop,ptr,UNDOXBYTE);
    undop->usbyte=0;
   }
  }

  undop->uend=xptr;
  undop->ustart=ptr;
 }
 else
 {
  len=getxword(undop,undop->ustart,UNDOXLAST);
  ptr=setptr(undop,undop->ustart+len);
  lptr=ptr;

  last=(len==sizeof(int));

 /* dprintf(0,"undo drop start=%d end=%d",undop->ustart,undop->uend); */ 

  while(!last)
  {
   len=getxword(undop,lptr-sizeof(int),UNDOXBYTE);
   lptr=setptr(undop,lptr-sizeof(int)-len);
 
   op=readword(undop,lptr);

 /* dprintf(2,"ulast=%d op=%x op-1=%x op1=%x",lptr,op,
               readword(undop,lptr-sizeof(int)),
               readword(undop,lptr+sizeof(int)));  */


   sptr=undop->uptr;
   undop->uptr=setptr(undop,lptr+sizeof(int));
   sstate=undop->state;
   undop->state=STUNDOSCRAP;

   last=op & UNDOXLAST;
   err=undoexecute(op & (~UNDOXLAST),file,1);
   undop=getundo(file);

   undop->uptr=sptr;
   undop->state=sstate;

   if(err)
   {
    dprintf(6,"undo drop bad op code *******"); 
    break;
   }
  }

  undop->ustart=ptr;
  len=readword(undop,ptr-sizeof(int));

  if(undop->ustart!=undop->uend)
  {
   if(len & UNDOXBYTE)
   {
    undop->usbyte=1;
    for(i=0;i<4;i++)
    {
     if(undop->data[ptr]) break;
     ptr=setptr(undop,ptr+1);
    }
    undop->ustart=ptr;
   }
  }

 /* dprintf(0,"undo drop start=%d end=%d",undop->ustart,undop->uend); */ 
 }

 *undopp=undop;

 return(NULL);
}



static os_error * undodroptop(filestr * file)
{
 os_error * err;
 undostr  * undop;
 int        len;
 int        last;
 int        op;
 int        i;
 int        ptr;

 err=NULL;
 undop=getundo(file);

 if(undop->ustart!=undop->uend)
 {
  undop->state=STUNDOSCRAP;
  undop->uostart=setptr(undop,undop->uend-sizeof(int));

  while(1)
  {
   len=getxword(undop,undop->uend-sizeof(int),UNDOXBYTE);
   undop->ulast=setptr(undop,undop->uend-sizeof(int)-len);
   undop->uptr=setptr(undop,undop->ulast+sizeof(int));
 
   op=readword(undop,undop->ulast);

   last=op & UNDOXLAST;
   undoexecute(op & (~UNDOXLAST),file,1);
   undop=getundo(file);
   undop->uend=undop->ulast;

   if(last) break;
  }
                                         
  undop->uend=setptr(undop,undop->ulast-sizeof(int));
  len=readword(undop,undop->uend); /* old start, will becomes new end */

  if(undop->ustart!=undop->uend)
  {
   if(len & UNDOXBYTE)
   {
    undop->uebyte=1;
    for(i=0;i<4;i++)
    {
     ptr=setptr(undop,undop->uend-1);
     if(undop->data[ptr]) break;
     else                 undop->uend=ptr;
    }
   }
  }

  undop->state=STNONE;
 }
 return(err);
}












static int canundodrop(undostr * undop)
{
 int len;
 int diff;

 if(undop->state==STCHANGE) return(undop->uend!=undop->ustart);
 else
 if(undop->state==STNEW || undop->state==STREDO)
 {
  return(undop->ustart!=undop->uostart);
 }
 else
 {
  if(undop->uend==undop->ustart) return(0);
  else
  if(!undop->usbyte)
  {
   len=getxword(undop,undop->ustart,UNDOXBYTE);
   diff=diffp(undop,undop->uend,undop->ustart);
   return(diff>len);
  }
  else return(1);
 }
}



/* drop an item from the end of the redo queue */

static os_error * redodrop(filestr * file,undostr ** undopp)
{
 os_error * err;
 undostr  * undop;
 int        len;
 int        op;
 int        last;
 int        ptr;
 int        xptr;
 int        sptr;
 int        lptr;
 int        sstate;
 int        i;


 undop=*undopp;

/* dprintf(0,"redo drop byte=%d",undop->rebyte);
 dprintf(1,"redo drop start=%d end=%d",undop->rstart,undop->rend);  */

 if(undop->rebyte)
 {
  ptr=undop->rend;
  xptr=undop->rstart;

  for(i=0;i<BYTECHUNK;i++)
  {
   ptr=setptr(undop,ptr-1);
   if(ptr==xptr) break;
   if(!undop->data[ptr]) break;    /* dissimilar to fix in undodrop */
  }

 /* dprintf(2,"ptr=%d xptr=%d",ptr,xptr);  */

  if(i<BYTECHUNK)
  {
   if(ptr==xptr)
   {
    xptr=ptr=rounddown(undop,ptr);
   }
   else
   {
    ptr=rounddown(undop,ptr);
    bicword(undop,ptr-sizeof(int),UNDOXBYTE);
    undop->rebyte=0;
   }
  }

  undop->rstart=xptr;
  undop->rend=ptr;
 }
 else
 {
  len=getxword(undop,undop->rend-sizeof(int),UNDOXLAST);
  ptr=setptr(undop,undop->rend-len);
  lptr=ptr;

  last=(len==sizeof(int));

 /* dprintf(0,"redo drop (1) start=%d end=%d",undop->rstart,undop->rend);
  dprintf(1,"(3) len=%d",len);  */

  while(!last)
  {
   len=getxword(undop,lptr,UNDOXBYTE);
   lptr=setptr(undop,lptr+len);

   op=readword(undop,lptr);

   sptr=undop->rptr;
   undop->rptr=lptr;
   sstate=undop->state;
   undop->state=STREDOSCRAP;

   last=op & UNDOXLAST;

   err=undoexecute(op & (~UNDOXLAST),file,1);
   undop=getundo(file);

   undop->rptr=sptr;
   undop->state=sstate;

   lptr=setptr(undop,lptr+sizeof(int));

   if(err)
   {
    dprintf(6,"redo drop bad op code *******"); 
    break;
   }
  }

 /* dprintf(0,"redo drop (2) start=%d end=%d",undop->ustart,undop->uend); */

  undop->rend=ptr;

  if(undop->rend!=undop->rstart)
  {
   len=readword(undop,ptr);

   if(len & UNDOXBYTE)
   {
    undop->rebyte=1;
    for(i=0;i<4;i++)
    {
     if(readbyte(undop,ptr-1)) break;
     else ptr=setptr(undop,ptr-1);
    }
    undop->rend=ptr;
   }
  }
 }

 *undopp=undop;

/* dprintf(2,"redo x drp start=%d end=%d",undop->rstart,undop->rend); */

 return(NULL);
}



static os_error * redodroptop(filestr * file)
{
 os_error * err;
 undostr  * undop;
 int        len;
 int        last;
 int        op;
 int        i;

 err=NULL;
 undop=getundo(file);
 if(undop->rstart!=undop->rend)
 {
  undop->state=STREDOSCRAP;

  undop->rostart=undop->rstart;

  while(1)
  {
   len=getxword(undop,undop->rstart,UNDOXBYTE);
   undop->rlast=setptr(undop,undop->rstart+len);
   undop->rptr=undop->rlast;

   op=readword(undop,undop->rlast);

   last=op & UNDOXLAST;
   undoexecute(op & (~UNDOXLAST),file,1);

   undop=getundo(file);
   undop->rstart=setptr(undop,undop->rlast+sizeof(int));

   if(last) break;
  }

  len=readword(undop,undop->rstart);
  undop->rstart=setptr(undop,undop->rstart+sizeof(int));

  if(len & UNDOXBYTE)
  {
   undop->rsbyte=1;
   for(i=0;i<3;i++) if(readbyte(undop,undop->rstart++)) break;
  }

  undop->state=STNONE;
 }

 return(err);
}



static int canredodrop(undostr * undop)
{
 int len;
 int diff;


 if(undop->state==STCHANGE) return(undop->rend!=undop->rstart);
 else
 if(undop->state==STUNDO)
 {
  return(undop->rend!=undop->rostart);   /* probably wrong */
 }
 else
 {
  if(undop->rend==undop->rstart) return(0);
  else
  if(!undop->rebyte)
  {
   len=getxword(undop,undop->rend-sizeof(int),UNDOXBYTE);
   diff=diffp(undop,undop->rend,undop->rstart);
   return(diff>len);
  }
  else return(1);
 }
}


/* ensures that space bytes are available to extend the undo stack upwards */
/* much easier when redo not active, and in general should only need to work */
/* in that case */

static os_error * undospace(filestr * file,undostr ** undopp,int space)
{
 undostr * undop;
 int       usize;
 int       rsize;
 int       maxsize;

 undop=*undopp;

 maxsize=undop->maxsize;

 usize=diffp(undop,undop->uend,undop->ustart);
 rsize=diffp(undop,undop->rend,undop->rstart);

/* dprintf(7,"us=%d rs=%d max=%d space=%d",usize,rsize,maxsize,space);*/


 while(((usize+rsize+space)>=maxsize) && canundodrop(undop))
 {
  undodrop(file,&undop);
  usize=diffp(undop,undop->uend,undop->ustart);
  rsize=diffp(undop,undop->rend,undop->rstart);

 /* dprintf(8,"drop us=%d rs=%d max=%d space=%d",usize,rsize,maxsize,space);*/
 }

 while(((usize+rsize+space)>=maxsize) && canredodrop(undop))
 {
  redodrop(file,&undop);
  usize=diffp(undop,undop->uend,undop->ustart);
  rsize=diffp(undop,undop->rend,undop->rstart);
 }

 if((usize+rsize+space)>=maxsize) undop->uscrap=1;

 *undopp=undop;

 return(NULL);
}



/* ensures that space bytes are available to extend the redo stack downwards */
/* much easier when redo not active, and in general should only need to work */
/* in that case */

static os_error * redospace(filestr * file,undostr ** undopp,int space)
{
 undostr * undop;
 int       usize;
 int       rsize;
 int       maxsize;

 undop=*undopp;

 maxsize=undop->maxsize;

 usize=diffp(undop,undop->uend,undop->ustart);
 rsize=diffp(undop,undop->rend,undop->rstart);

 while(((usize+rsize+space)>=maxsize) && canredodrop(undop))
 {
  redodrop(file,&undop);
  usize=diffp(undop,undop->uend,undop->ustart);
  rsize=diffp(undop,undop->rend,undop->rstart);
 }

 while(((usize+rsize+space)>=maxsize) && canundodrop(undop))
 {
  undodrop(file,&undop);
  usize=diffp(undop,undop->uend,undop->ustart);
  rsize=diffp(undop,undop->rend,undop->rstart);
 }

 if((usize+rsize+space)>=maxsize) undop->rscrap=1;

 *undopp=undop;

 return(NULL);
}


/* used to empty the redo buffer */

static os_error * redotrash(filestr * file,undostr ** undopp)
{
 undostr * undop;

 undop=*undopp;

 while(undop->rend!=undop->rstart) 
 {
  redodrop(file,&undop);
 }

 *undopp=undop;

 return(NULL);
}





#ifdef NEVER

/* used to empty the undo buffer */

static os_error * undotrash(filestr * file,undostr ** undopp)
{
 undostr * undop;

 undop=*undopp;

 while(undop->uend!=undop->ustart) 
 {
  undodrop(file,&undop);
 }

 *undopp=undop;

 return(NULL);
}


static os_error * emptyundobuffer(filestr * file)
{
 undostr * undop;

 undop=getundo(file);

 redotrash(file,&undop);
 undotrash(file,&undop);
 return(NULL);
}

#endif





os_error * undowrite(filestr * file,void * data,int size)
{
 undostr       * undop;
 int             len;
 int             extra;

/* dprintf(0,"undowrite in"); */

 if(size)
 {
  undop=getundo(file); 

  if(undop->state==STREDO || undop->state==STNEW)
  {
   if(undop->state==STNEW) extra=MINUNDOSPACE+REDOGAP;
   else                    extra=MINUNDOSPACE;

   if(!undop->uscrap) undospace(file,&undop,size+extra);

   if(!undop->uscrap)
   {
    if((undop->uptr+size)>=undop->maxsize)
    {
     len=undop->maxsize-undop->uptr;
     if(len>0)
     {
      XMEMCPYX(undop->data+undop->uptr,data,len);
      size-=len;
      data=((char*)data)+len;
     }
     undop->uptr=0;
    }

    XMEMCPYX(undop->data+undop->uptr,data,size);
    undop->uptr+=size;

    undop->uend=undop->uptr;
   }
  }
  else
  if(undop->state==STUNDO)
  {
   if(!undop->rscrap) redospace(file,&undop,size+MINUNDOSPACE);

   if(!undop->rscrap)
   {
    if((undop->rptr-size)<0)
    {
     len=undop->rptr;
     XMEMCPYX(undop->data,(char*)data+size-len,len);
     size-=len;
     undop->rptr=undop->maxsize;
    }
    XMEMCPYX(undop->data+undop->rptr-size,data,size);
    undop->rptr-=size;

    undop->rstart=undop->rptr;
   }
  }
 }

/* dprintf(0,"undowrite ex"); */

 return(NULL);
}



static void writeredofn(undostr * undop)
{
 int len;

 if(undop->rwfn)
 {
  len=diffp(undop,undop->rlast,undop->rptr-sizeof(int));
  undop->rptr=setptr(undop,undop->rptr-sizeof(int));
  writeword(undop,undop->rptr,len);
 }
}


static void writeundofn(undostr * undop) 
{
 int len;

 if(undop->uwfn)
 {
  len=diffp(undop,undop->uptr,undop->ulast);
 /* dprintf(0,"len=%d at %d",len,undop->uptr);  */
  writeword(undop,undop->uptr,len);
  undop->uptr=setptr(undop,undop->uptr+sizeof(int));
 }
}



os_error * undowritefn2(filestr * file,int fn,void * data,int size1,int size2)
{
 undostr       * undop;
 int             sstate;
 int             extra;

/* dprintf(0,"undowrite2 in"); */

 undop=getundo(file);

 if(undop->state==STREDO || undop->state==STNEW)
 {
  if(undop->state==STNEW) extra=MINUNDOSPACE+REDOGAP;
  else                    extra=MINUNDOSPACE;


  if(!undop->uscrap)    /* theory 2*sizeof(int) allow for rounding */
     undospace(file,&undop,size1+size2+4*sizeof(int)+extra);

  if(!undop->uscrap)
  {
   undop->uptr=roundup(undop,undop->uptr);
   writeundofn(undop);
   undop->ulast=undop->uptr;
   undowrite(file,&fn,sizeof(int));
   undowrite(file,data,size1);
   if(size2)
   {
    /* data=((char*)data)+size1; */
    undowrite(file,((char*)data)+size1,size2);
   }
   undop->uwfn=1;
  }     /* if scrap is now true, you are potentially in trouble */

  if(undop->uscrap)
  {
   sstate=undop->state;
   fastptr=(char*)data;
   fastmax=size1+size2;
   undop->state=STFASTSCRAP;
   undoexecute(fn,file,1);
   undop=getundo(file);
   undop->state=sstate;
  }
 }
 else
 if(undop->state==STUNDO)
 {
  if(!undop->rscrap)
     redospace(file,&undop,size1+size2+2*sizeof(int)+MINUNDOSPACE);

  if(!undop->rscrap)
  {
   undop->rptr=rounddown(undop,undop->rptr);
   writeredofn(undop);
   undop->rlast=setptr(undop,undop->rptr-sizeof(int));
   undowrite(file,&fn,sizeof(int));
   undowrite(file,data,size1);
   if(size2)
   {
    /* data=((char*)data)+size1; */
    undowrite(file,((char*)data)+size1,size2);
   }
   undop->rwfn=1;
  }

  if(undop->rscrap)
  {
   sstate=undop->state;
   fastptr=(char*)data;
   fastmax=size1+size2;
   undop->state=STFASTSCRAP;
   undoexecute(fn,file,1);
   undop=getundo(file);
   undop->state=sstate;
  }
 }


/* dprintf(0,"undowrite2 ex"); */

 return(NULL);
}



os_error * undowritefn(filestr * file,int fn,void * data,int size)
{
 return(undowritefn2(file,fn,data,size,0));
}





int undoscrapcheck(filestr * file,int size)
{
 undostr       * undop;
 int             extra;


 undop=getundo(file);

 if(undop->state==STREDO || undop->state==STNEW)
 {
  if(undop->state==STNEW) extra=MINUNDOSPACE+REDOGAP;
  else                    extra=MINUNDOSPACE;

  if(!undop->uscrap)    /* theory 2*sizeof(int) allow for rounding */
     undospace(file,&undop,size+4*sizeof(int)+extra);

  return(undop->uscrap);
 }
 else
 if(undop->state==STUNDO)
 {
  if(!undop->rscrap)
     redospace(file,&undop,size+2*sizeof(int)+MINUNDOSPACE);

  return(undop->rscrap);
 }

 return(0);
}









os_error * undoread(filestr * file,void * data,int size)
{
 undostr       * undop;
 int             len;

 undop=getundo(file);

 if(undop->state==STFASTSCRAP)
 {
  if(fastmax>=size)
  {
   XMEMCPYX(data,fastptr,size);
   fastptr+=size;
   fastmax-=size;
  }
 }
 else
 {
  if(undop->state==STUNDO || undop->state==STUNDOSCRAP)
  {
   if((undop->uptr+size)>=undop->maxsize)
   {
    len=undop->maxsize-undop->uptr;
    XMEMCPYX(data,undop->data+undop->uptr,len);
    size-=len;
    undop->uptr=0;
    data=((char*)data)+len;
   }

   XMEMCPYX(data,undop->data+undop->uptr,size);
   undop->uptr+=size;
  }
  else
  if(undop->state==STREDO || undop->state==STREDOSCRAP)
  {
   if((undop->rptr-size)<0)
   {                                                                                len=undop->rptr;
    XMEMCPYX((char*)data+size-len,undop->data,len);
    size-=len;
    undop->rptr=undop->maxsize;
   }
   XMEMCPYX(data,undop->data+undop->rptr-size,size);
   undop->rptr-=size;
  }
 }

 return(NULL);
}





static os_error * xundostart(filestr * file,int state)
{
 undostr * undop;

 undop=getundo(file);

 if(undop->state==STNONE)
 {
  undop->count=0;

  undop->uscrap=(undop->maxsize<=0);

  undop->state=state;

  if(undop->ustart==undop->uend)
  {
   undop->uewbyte=undop->uebyte=undop->usbyte=0;
  }
  else
  if(undop->uebyte)
  {
   do
   {
    writebyte(undop,undop->uend,0);
    undop->uend=setptr(undop,undop->uend+1);
   } while(undop->uend & 0x3);
   undop->uewbyte=undop->uebyte;
   undop->uebyte=0;
  }

  undop->uend=roundup(undop,undop->uend);
  undop->uostart=undop->uend;
  undop->ulast=undop->uptr=setptr(undop,undop->uend+sizeof(int));
  undop->uwfn=0;

  if(state==STNEW) redotrash(file,&undop); /* trash redo buffer */
 }
 else
 {
  undop->count++;
 }

 return(NULL);
}


static os_error * xundoend(filestr * file,int state)
{
 undostr * undop;
 int       len;

 undop=getundo(file);

 if(!undop->count--)
 {
  if((int)undop->state==state)
  {
   if(undop->uwfn)
   {
    /* this is standard end to operation */

    undop->uptr=roundup(undop,undop->uptr);
                                                /* mark first op code as end */
    orrword(undop,undop->uostart+sizeof(int),UNDOXLAST); 
    writeundofn(undop);
    
    len=diffp(undop,undop->uptr,undop->uostart);
    writeword(undop,undop->uostart,len);
    if(undop->uewbyte) orrword(undop,undop->uostart,UNDOXBYTE);

    undop->uend=undop->uptr;

    undop->uewbyte=0;

    if(undop->uscrap) /* but if a write failed, retrieve position */
    {
     /* undop->uend=undop->uostart; */

 /* dprintf(0,"undo drop start=%d end=%d",undop->ustart,undop->uend); */
     undodrop(file,&undop);
    }
   }
   undop->state=STNONE;
  }
 }

 return(NULL);
}


os_error * undostart(filestr * file)
{
 return(xundostart(file,STNEW));
}

os_error * undoend(filestr * file)
{
 return(xundoend(file,STNEW));
}



os_error * undobyte(filestr * file,int c)
{
 undostr       * undop;
 int             extra;


 undop=getundo(file);

 if(undop->state==STUNDO)
 {
  if(undop->rend==undop->rstart)
  {
   undospace(file,&undop,REDOGAP+MINUNDOSPACE);

   undop->rstart=undop->rend=setptr(undop,(undop->uend & (~0x3))+REDOGAP);
   undop->rsbyte=undop->rebyte=1;
   undop->rswbyte=0;
   undop->rstart=undop->rend=setptr(undop,undop->rstart-1);
   writebyte(undop,undop->rstart,0);
  }
  else
  {
   redospace(file,&undop,MINUNDOSPACE);

   if(!undop->rsbyte)
   {
    orrword(undop,undop->rstart,UNDOXBYTE);
    undop->rstart=setptr(undop,undop->rstart-1);
    writebyte(undop,undop->rstart,0);
    undop->rsbyte=1;
   }
  }

  undop->rstart=setptr(undop,undop->rstart-1);
  writebyte(undop,undop->rstart,c);
 }
 else
 {
  if(undop->state!=STREDO) redotrash(file,&undop);

  if(undop->state==STNEW) extra=MINUNDOSPACE+REDOGAP;
  else                    extra=MINUNDOSPACE;

  undospace(file,&undop,extra);

  if(!undop->uebyte)
  {
   if(undop->ustart!=undop->uend)
                             orrword(undop,undop->uend-sizeof(int),UNDOXBYTE);
   writebyte(undop,undop->uend,0);
   undop->uend=setptr(undop,undop->uend+1);
   undop->uebyte=1;
  }

  writebyte(undop,undop->uend,c);
  undop->uend=setptr(undop,undop->uend+1);
 }

 return(NULL);
}


static os_error * redostart(filestr * file)
{
 undostr * undop;

 undop=getundo(file);

 undop->rscrap=(undop->maxsize<=0);
 if(!undop->rscrap)
 {
  if(undop->rend==undop->rstart)
  {
   undospace(file,&undop,REDOGAP+MINUNDOSPACE);
   undop->rstart=undop->rend=setptr(undop,(undop->uend & (~0x3))+REDOGAP);
   undop->rsbyte=undop->rswbyte=undop->rebyte=0;
  }
  else
  {
   if(undop->rsbyte)
   {
    do
    {
     undop->rstart=setptr(undop,undop->rstart-1);
     undop->data[undop->rstart]=0;
    } while(undop->rstart & 0x3);

    undop->rswbyte=undop->rsbyte;
    undop->rsbyte=0;
   }

   undop->rstart=rounddown(undop,undop->rstart);
  }

  undop->rostart=setptr(undop,undop->rstart-sizeof(int));
  undop->rptr=undop->rlast=undop->rostart;
  undop->rwfn=0;
 }

 return(NULL);
}


static os_error * redoend(filestr * file)
{
 undostr * undop;
 int       len;

 undop=getundo(file);

 if(undop->state==STUNDO)
 {
  if(undop->rwfn)
  {
   undop->rptr=rounddown(undop,undop->rptr);
                                           /* mark last op code as end */
   orrword(undop,undop->rostart-sizeof(int),UNDOXLAST); 
   writeredofn(undop);

   len=diffp(undop,undop->rostart+sizeof(int),undop->rptr);
   writeword(undop,undop->rostart,len);
   if(undop->rswbyte) orrword(undop,undop->rostart,UNDOXBYTE);
   undop->rstart=undop->rptr;
   /* dprintf(6,"redo end ptr=%d start=%d",undop->rstart,undop->rptr); */
   undop->rswbyte=0;

   if(undop->rscrap)  /* a write failed, retrive */
   {
    /* undop->rstart=setptr(undop,undop->rostart+sizeof(int)); */
    redodrop(file,&undop);
   }
  }
 }
 return(NULL);
}



/* read from undo, write to redo */

static os_error * xbyteundo(undostr * undop,filestr * file)
{
 int byte;
 int ptr;

 undop->state=STUNDO;

 /* do something with the byte */

 undop->uend=setptr(undop,undop->uend-1);
 byte=readbyte(undop,undop->uend);

 if(undop->ustart!=undop->uend)
 {
  ptr=setptr(undop,undop->uend-1);
  if(!undop->data[ptr])
  {
   undop->uend=rounddown(undop,undop->uend);
   undop->uebyte=0;
  }
 }

 if(undosb) undosb(byte,file);

 undop=getundo(file);
 undop->state=STNONE;

 return(NULL);
}


static os_error * xundo(filestr * file)
{
 os_error * err;
 undostr  * undop;
 int        len;
 int        last;
 int        op;
 int        i;
 int        ptr;

 err=NULL;
 undop=getundo(file);

/* dprintf(1,"ustart=%d uend=%d",undop->ustart,undop->uend); */

 if(undop->ustart!=undop->uend)
 {
  if(undop->uebyte) err=xbyteundo(undop,file);
  else
  {
   undop->state=STUNDO;
   undop->uostart=setptr(undop,undop->uend-sizeof(int));
   redostart(file);

   while(1)
   {
    len=getxword(undop,undop->uend-sizeof(int),UNDOXBYTE);
    undop->ulast=setptr(undop,undop->uend-sizeof(int)-len);
    undop->uptr=setptr(undop,undop->ulast+sizeof(int));
 
    op=readword(undop,undop->ulast);

  /*  dprintf(5,"op=%x",op);  */

    last=op & UNDOXLAST;
    undoexecute(op & (~UNDOXLAST),file,0);
    undop=getundo(file);
    undop->uend=undop->ulast;

    if(last) break;
   }
                                         
   undop->uend=setptr(undop,undop->ulast-sizeof(int));
   len=readword(undop,undop->uend); /* old start, will becomes new end */

   if(undop->ustart!=undop->uend)
   {
    if(len & UNDOXBYTE)
    {
     undop->uebyte=1;
     for(i=0;i<4;i++)
     {
      ptr=setptr(undop,undop->uend-1);
      if(undop->data[ptr]) break;
      else                 undop->uend=ptr;
     }
    }
   }

   redoend(file);
   undop->state=STNONE;
  }
 }

 return(err);
}


/* read from redo, write to undo */

static os_error * xbyteredo(undostr * undop,filestr * file)
{
 int byte;
 int byte2;

 undop->state=STREDO;

 /* do something with the byte */

 byte=readbyte(undop,undop->rstart);
 undop->rstart=setptr(undop,undop->rstart+1);
 if(undop->rstart!=undop->rend)
 {
  byte2=readbyte(undop,undop->rstart);
  if(!byte2)
  {
   undop->rstart=roundup(undop,undop->rstart);
   undop->rsbyte=0;
  }
 }

 if(undosb) undosb(byte,file);

 undop=getundo(file);
 undop->state=STNONE;

 return(NULL);
}


static os_error * xredo(filestr * file)
{
 os_error * err;
 undostr  * undop;
 int        len;
 int        last;
 int        op;
 int        i;

 err=NULL;
 undop=getundo(file);
 if(undop->rstart!=undop->rend)
 {
/* dprintf(1,"rend=%d rstart=%d",undop->rend,undop->rstart); 
 dprintf(1,"ustart=%d uend=%d",undop->ustart,undop->uend); */


  if(undop->rsbyte) err=xbyteredo(undop,file);
  else
  {
   undop->rostart=undop->rstart;
   xundostart(file,STREDO);

   while(1)
   {
    len=getxword(undop,undop->rstart,UNDOXBYTE);
    undop->rlast=setptr(undop,undop->rstart+len);
    undop->rptr=undop->rlast;

    op=readword(undop,undop->rlast);
   /* dprintf(0,"redo op=%x",op); */

    last=op & UNDOXLAST;
    undoexecute(op & (~UNDOXLAST),file,0);

    undop=getundo(file);
    undop->rstart=setptr(undop,undop->rlast+sizeof(int));

    if(last) break;
   }

   len=readword(undop,undop->rstart);
   undop->rstart=setptr(undop,undop->rstart+sizeof(int));

   if(len & UNDOXBYTE)
   {
    undop->rsbyte=1;
    for(i=0;i<3;i++) if(readbyte(undop,undop->rstart++)) break;
   }

   xundoend(file,STREDO);
  }
 }

/* dprintf(1,"rend=%d rstart=%d",undop->rend,undop->rstart); 
 dprintf(1,"ustart=%d uend=%d",undop->ustart,undop->uend); */

 return(err);
}


/* return flags for undo and redo */

int canundos(filestr * file)
{
 undostr * undop;
 undop=getundo(file);

 if(!undop) return(0);

 return(undop->ustart!=undop->uend);
}


int canundo(int fp)
{
 return(canundos(menufile));
 USE(fp);
}


int undostring(int fp)
{
 char      p[MAXUNDOSTRING];
 char    * q;
 undostr * undop;
 int       op;
 int       ptr;
 int       len;
 int       c;

 p[0]=0;

 undop=getundo(menufile);
 if(undop->ustart!=undop->uend)
 {
  if(undop->uebyte)
  {
   if(undosbname)
   {
    c=readbyte(undop,undop->uend-1);
    q=undosbname(c);
    if(q) strcpy(p,q);
   }
  }
  else
  {
   len=getxword(undop,undop->uend-sizeof(int),UNDOXBYTE);
   ptr=setptr(undop,undop->uend-len-sizeof(int));
   op=getxword(undop,ptr,UNDOXLAST);
   getopstring(op,p);
  }
  trans(p,MAXUNDOSTRING);
 }

 return(createstring(p));

 USE(fp);
}



int canredo(int fp)
{
 undostr * undop;
 undop=getundo(menufile);

 if(!undop) return(0);

 return(undop->rstart!=undop->rend);
 USE(fp);
}



int redostring(int fp)
{
 char      p[MAXUNDOSTRING];
 char    * q;
 undostr * undop;
 int       op;
 int       ptr;
 int       len;
 int       c;

 p[0]=0;

 undop=getundo(menufile);
 if(undop->rstart!=undop->rend)
 {
  if(undop->rsbyte)
  {
   if(undosbname)
   {
    c=readbyte(undop,undop->rstart);
    q=undosbname(c);
    if(q) strcpy(p,q);
   }
  }
  else
  {
   len=getxword(undop,undop->rstart,UNDOXBYTE);
   ptr=setptr(undop,undop->rstart+len);
   op=getxword(undop,ptr,UNDOXLAST);
   getopstring(op,p);
  }
  trans(p,MAXUNDOSTRING);
 }

 return(createstring(p));

 USE(fp);
}





static os_error * undotestfile(filestr * file)
{
 undostr  * undop;

 undop=getundo(file);
 if(undop)
 {
  if(undop->maxsize==0)  /* some files have 0 maxsize but pointers */
  {
/* dprintf(1,"rend=%d rstart=%d",undop->rend,undop->rstart); 
 dprintf(2,"ustart=%d uend=%d",undop->ustart,undop->uend); */


   undop->rstart=undop->rend=undop->rlast=0;
   undop->ustart=undop->uend=undop->ulast=0;
  }
 }
 return(NULL);
}



os_error * undosx(viewstr * view,filestr * file)
{
 undotestfile(file);

 checktools(view,file);
 xundo(file);
 return(redodroptop(file));
}


void undo(int fp)
{
 checktools(menuview,menufile);
 zreport(xundo(menufile));
 USE(fp);
}

os_error * undov(viewstr * view)
{
 checktools(view,view->file);
 return(xundo(view->file));
}


void redo(int fp)
{
 checktools(menuview,menufile);
 zreport(xredo(menufile));
 USE(fp);
}


os_error * redov(viewstr * view)
{
 checktools(view,view->file);
 return(xredo(view->file));
}


/*****************************************************************************/

static void undoreloc(undostr * undop,int at,int by)
{
 if(undop->ustart>=at)  undop->ustart+=by;
 if(undop->uend>=at)    undop->uend+=by;
 if(undop->ulast>=at)   undop->ulast+=by;
 if(undop->uptr>=at)    undop->uptr+=by;
 if(undop->uostart>=at) undop->uostart+=by;

 if(undop->rstart>=at)  undop->rstart+=by;
 if(undop->rend>=at)    undop->rend+=by;
 if(undop->rlast>=at)   undop->rlast+=by;
 if(undop->rptr>=at)    undop->rptr+=by;
 if(undop->rostart)     undop->rostart+=by;
}


static void undocompact(undostr * undop,int start,int end,int old,int xnew)
{
 if(start<end && old>xnew)
 {
  memmove(undop->data,undop->data+start,end-start);
  undoreloc(undop,start,-start);
 }
 else
 if(start>end && old!=xnew)
 {
  memmove(undop->data+start-(old-xnew),undop->data+start,old-start);
  undoreloc(undop,start,-(old-xnew));
 }
}


static void undoimpact(undostr * undop,int start,int end,int old,int xnew)
{
 int shift;

 if(start<end && old>xnew)
 {
  shift=xnew-old;
  memmove(undop->data+end-shift,undop->data+end,old-end);
  undoreloc(undop,end,-shift);
 }
 else
 if(start>end && old!=xnew)
 {
  memmove(undop->data,undop->data+end,end);
  undoreloc(undop,end,-end);
 }
}






static os_error * undosetfilesize(filestr * file,int size)
{
 os_error * err;
 undostr  * undop;
 int        maxsize;
 int        rsize;


/* dprintf(5,"undosetfilesize =%d",size); */

 err=NULL;

 undop=getundo(file);
 if(undop)
 {
  maxsize=undop->maxsize;

/*
  dprintf(4,"undo maxsize=%d size=%d",maxsize,size); 

  dprintf(6,"rend=%d rstart=%d uend=%d ustart=%d",
          undop->rend,undop->rstart,undop->uend,undop->ustart);   
*/

  if(maxsize==0)  /* some files have 0 maxsize but pointers */
  {
   undop->rstart=undop->rend=undop->rlast=0;
   undop->ustart=undop->uend=undop->ulast=0;
  }

  if(size!=maxsize)
  {
   undop->state=STCHANGE;

   if(size<maxsize)  /* making smaller */
   {
    rsize=maxsize-size;
    if(undop->rend==undop->rstart) rsize+=MINUNDOSPACE+REDOGAP;
    else                           rsize+=MINUNDOSPACE;

    undospace(file,&undop,rsize);


  /* dprintf(6,"rend=%d rstart=%d uend=%d ustart=%d",
          undop->rend,undop->rstart,undop->uend,undop->ustart); */

    if(undop->rend!=undop->rstart && undop->uend!=undop->ustart)
    {
     rsize=diffp(undop,undop->rstart,undop->uend);
     undocompact(undop,undop->ustart,undop->rend,maxsize,size+rsize);
     undoimpact(undop,undop->uend,undop->rstart,size+rsize,size);
    }
    else
    if(undop->uend!=undop->ustart)
    {
     undocompact(undop,undop->ustart,undop->uend,maxsize,size);
    }
    else
    if(undop->rend!=undop->rstart)
    {
     undocompact(undop,undop->rstart,undop->rend,maxsize,size);
    }

    err=fob_change(undop->key,file,size+sizeof(undostr)-sizeof(int),
                                                        (fobjptr)&undop);  
   }
   else                     /* making bigger  */
   if(size>maxsize)
   {
    err=fob_change(undop->key,file,size+sizeof(undostr)-sizeof(int),
                                                        (fobjptr)&undop);  

    if(!err)
    {
     if(undop->uend!=undop->ustart && undop->rend!=undop->rstart)
     {
      undocompact(undop,undop->ustart,undop->rend,maxsize,size);
     }
     else
     if(undop->uend!=undop->ustart)
     {
      undocompact(undop,undop->ustart,undop->uend,maxsize,size);
     }
     else
     if(undop->rend!=undop->rstart)
     {
      undocompact(undop,undop->rstart,undop->rend,maxsize,size);
     }
    }
   }

   undop->maxsize=size;
   undop->state=STNONE;
  }
 }

 return(err);
}



os_error * undozerosize(filestr * file)
{
 undosetfilesize(file,0);
 file->zeroundobuffer=1;
 fob_compact(file);
 return(NULL);
}



os_error * undosetsize(int size)
{
 filestr * file;

 undosize=size;

 file=firstfile;
 while(file)
 {
  if(!file->zeroundobuffer)
  {
   undosetfilesize(file,size);
  }
  file=file->next;
 }

 fob_compactall();

 return(NULL);
}



/* if save with no undo buffer, then trash the buffer and set size to zero */

os_error * undosavefile(filestr * file,int type)
{
 os_error * err;

 err=NULL;

 if((saveundo && undosize) || (type & FRSAVEUNDO))
 {
  undostart(file);
  toolsaveposn(file);
  event(EVENT_SAVEPOSN,file,0);
  undoend(file);
 }
 else
 {
  if(!file->zeroundobuffer)
  {
   err=undosetfilesize(file,0);
  }
 }

 return(err);

 USE(type);
}


/* if undo buffer has zero size, then set to correct size */

os_error * undopostsavefile(filestr * file)
{
 os_error * err;

 err=NULL;

 if(saveundo)
 {
  err=undodroptop(file);
 }
 else
 {
  if(!file->zeroundobuffer)
  {
   err=undosetfilesize(file,undosize);
  }
 }

 return(err);
}


/* this is our chance to make the undo buffer for this file the correct size */

os_error * undoloadfile(filestr * file)
{
 os_error * err;

 err=undosetfilesize(file,undosize);

 return(err);
}


void flushundobuffer(int fp)
{
 os_error * err;

          err=undosetfilesize(menufile,0);
 if(!err) err=undosetfilesize(menufile,undosize);
 if(!err) err=fob_compact(menufile);

 zreport(err);

 USE(fp);
}


/***************************************************************************/


#ifdef NEVER

os_error * undoaddsbfn(undosbfn sbfn,undosbnamefn sbnamefn)
{
 undosb=sbfn;
 undosbname=sbnamefn;
 return(NULL);
}

#endif



os_error * undoaddblock(undofnstr * undoblock,int size)
{
 os_error * err;

 err=flex_chunk((flex_ptr)&undos,(undon+size)*sizeof(undofnstr),UNDOCHUNK);

 if(!err)
 {
  XMEMCPYX(undos+undon,undoblock,size*sizeof(undofnstr));
  undon+=size;
 }
 return(err);
}



/* all undos have now been declared */
/* so sort the table                */


static os_error * undorun(int eventn,int userhandle,void * data,int data1)
{
 qsort(undos,undon,sizeof(undofnstr),undocmp);
 return(NULL);

 USE(eventn);
 USE(userhandle);
 USE(data);
 USE(data1);
}




static os_error * undomkstyle(int eventn,int user,filestr * file,
                                                    filestr * ofile)
{
 /* do nothing, style sheet has no undo buffer */

 return(NULL);

 USE(file);
 USE(ofile);
 USE(eventn);
 USE(user);
}



static os_error * undonewevent(int eventn,int user,filestr * file,
                                                             void * data)
{
 return(newundobuffer(file));
 USE(eventn);
 USE(data);
 USE(user);
}



static os_error * undobareevent(int eventn,int user,filestr * file,
                                                             void * data)
{
 return(newundobuffer(file));
 USE(eventn);
 USE(data);
 USE(user);
}



static os_error * undoloadstyle(int eventn,int user,filestr * file,
                                                             void * data)
{
 return(newundobuffer(file));
 USE(eventn);
 USE(data);
 USE(user);
}






/* called on program boot */

os_error * initundo(void)
{
 os_error * err;

 err=flex_alloc((flex_ptr)&undos,UNDOCHUNK);
 undon=0;
 if(!err) err=addevent(EVENT_RUNPROGRAM,undorun,0);
 if(!err) err=addevent(EVENT_MKSTYLE1,(eventfn)undomkstyle,0);
 if(!err) err=addevent(EVENT_NEWDOC1,(eventfn)undonewevent,0);
 if(!err) err=addevent(EVENT_BAREDOC1,(eventfn)undobareevent,0);
 if(!err) err=addevent(EVENT_LOADSTYLE1,(eventfn)undoloadstyle,0);

 return(err);
}

